<html lang="en">
  <head>
    <meta charset="utf-8"></meta>
    <title> {{title}} </title>
    <link rel="stylesheet" href="https://cdnjs.cloudflare.com/ajax/libs/bulma/0.6.0/css/bulma.min.css" media="all" />
    <link rel="shortcut icon" href="{{ rel_path_statics }}statics/images/favicon.ico"/>
  </head>

  <body>
  {% block content %} {% end %}


  </body>
</html>
